/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */

#ifdef _USRDLL
  #define _QAX_
#endif

//#include "notemplatewarning.h"

#include "qubaxo.h"

#include <math.h>
#include <stdexcept>
#include <iostream>
#include <sstream>
#include <string>
#include <string.h>
#include <vector>

#include <sstream>
#include <fstream>
using namespace std;

#include "Abffiles.h"
#include <acquirefile_read.h>
#include <QUB_Tree.h>

const long QAX_COPYING_MAXBYTES=(1<<15);	//TODO - test with 2^16 bytes - faster ? Smaller size should allow it to use preread buffer.

class callbk_reportbuf : public std::streambuf
{
public:
	callbk_reportbuf() : lineout( new ostringstream )
	{ setp(0,0), setg(0,0,0); }

	~callbk_reportbuf() { delete lineout; }

	void setTarget( int (__stdcall *cb)(const char*, void*), void *data )
	{
		callback = cb;
		this->data = data;
	}

protected:
	virtual int overflow( int ch )
{
	if ( ch != EOF ) {
		if ( ch == '\n' ) {
			if ( callback )
				callback( lineout->str().c_str(), data );
			delete lineout;
			lineout = new ostringstream;
		}
		else {
			(*lineout) << (char) ch;
		}
	}
	return ch;
}

private:
	int (__stdcall *callback)(const char *, void*);
	void *data;
	ostringstream *lineout;
};

class callbk_reportstream : public std::ostream
{
public:
	callbk_reportstream( int (__stdcall *cb)(const char *, void*), void *data )
		: std::basic_ostream<char,std::char_traits<char> >( new callbk_reportbuf )
	{
		((callbk_reportbuf *) rdbuf())->setTarget( cb, data );
	}
    virtual ~callbk_reportstream()
    {
            delete rdbuf();
	}
};

void WriteABFErr(int nError, const char *fname, ostream& out)
{
	char bufErr[256];
	ABF_BuildErrorText( nError, fname, bufErr, 255 );
	out << "ABFFIO " << bufErr << endl;
}

extern "C"
QAX_API int   __stdcall  Convert_ABF_To_QDF( const char *in_path, const char *out_path,
											 int (__stdcall *write_msg)(const char *msg, void *data),
											 void *data )
{
	// wrap the text-output callback in a std::ostream:
	callbk_reportstream msgout(write_msg, data);

	int hFile = 0;
	int nError;
	unsigned int iMaxSamples;
	DWORD dwMaxEpi;
	ABFFileHeader abfFH;
//		= {1.2, 3, 4, 5, 6, 7, 8, 9, 10.11, 12,
//		1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, {17, 18, 19, 20}, {33, 34, 35, 36},
//		1, 2.3, 4, 1, 6.7, 8.9, 10, 11, 12, 13, 14, 15, 16, 17, 18.19, 20, 21, 22, 23.24, 25.26, 27.28, 29.30, 31, 32, 33.34,
//		1, 2, 3, 4, 5, 6, 7.8, 9, 10,
//		1.2, 3.4, 5, 6, 7, 8, 9, 10, 11,
//		1, 2, 3.4, 5.6, 7.8};

//	ofstream hdr_out;
//	hdr_out.open("abfhdr");
//	hdr_out.write((const char *) &abfFH, sizeof(abfFH));
//	hdr_out.flush();
//	hdr_out.close();

	if ( ! ABF_ReadOpen( in_path, &hFile, ABF_DATAFILE, &abfFH, &iMaxSamples, &dwMaxEpi, &nError ) ) {
		WriteABFErr(nError, in_path, msgout);
		return -1;
	}

	// measure, set up buffer
	UINT iSamples = 1, iSampRead;
	for ( DWORD dwEpi = 1; dwEpi <= dwMaxEpi; ++dwEpi ) {
		ABF_GetNumSamples( hFile, &abfFH, dwEpi, &iSampRead, &nError );
		iSamples = max(iSamples, iSampRead);
	}
	vector<float> vSamples(iSamples);
	float *samples = & (vSamples[0]);

	// setup qdf
	QUB_Tree qdf = QUB_Tree::Create("DataFile");
	if ( ! qdf.saveAs(out_path) ) {
		msgout << "Failed to write " << out_path << endl;
		ABF_Close( hFile, NULL );
		return -2;
	}
	qdf["Sampling"].setData(QTR_TYPE_DOUBLE, abfFH.fADCSequenceInterval/1000000.0);
	qdf["Scaling"].setData(QTR_TYPE_DOUBLE, 1.0);
	qdf["ADDataSize"].setData(QTR_TYPE_INT, 2);
	qdf["ADDataType"].setData(QTR_TYPE_INT, (int) QTR_TYPE_SHORT);

	// set up channels
	int iChannels = abfFH.nADCNumChannels;
	vector<double> scaling_ch, offset_ch;
	double lo, hi, scaling, offset;
	QUB_TreeIter insChan = qdf["SessionFile"]["DataChannels"].children();
	for (int iCh=0; iCh<iChannels; ++iCh) {
		int jCh = abfFH.nADCSamplingSeq[iCh];
		insChan.insert("Channel");
		QUB_Tree chan = *insChan;
		++insChan;

		chan["Name"].setData( string(abfFH.sADCChannelName[jCh], 10) );
		chan["Units"].setData( string(abfFH.sADCUnits[jCh], 8) );
		
		// find dynamic range for scaling
		lo = 1e20;
		hi = -1e20;
		for ( DWORD dwEpi=1; dwEpi<=dwMaxEpi; ++dwEpi ) {
			if ( ! ABF_ReadChannel( hFile, &abfFH, jCh, dwEpi, samples, &iSampRead, &nError ) ) {
				WriteABFErr( nError, in_path, msgout );
				//return -3;
			}
			for (unsigned int i=0; i<iSampRead; ++i) {
				lo = min(lo, samples[i]);
				hi = max(hi, samples[i]);
			}
		}
		offset = 0;
		scaling = 32767.0 / max(abs(hi), abs(lo));
		scaling_ch.push_back(scaling);
		offset_ch.push_back(offset);
		chan["Scaling"].setData(QTR_TYPE_DOUBLE, scaling);
		chan["Offset"].setData(QTR_TYPE_DOUBLE, offset);
	}

	int iTotalChannels = iChannels;
	vector<int> ixDigi;
	bool keep;
	for (int iCh=0; iCh<ABF_DACCOUNT; ++iCh) {
		if ( abfFH.nWaveformEnable[iCh] ) {
			// find dynamic range for scaling (and also whether there's no waveform (!))
			lo = 1e20;
			hi = -1e20;
			keep = true;
			for ( DWORD dwEpi=1; dwEpi<=dwMaxEpi; ++dwEpi ) {
				if ( ! ABF_GetWaveform( hFile, &abfFH, iCh, dwEpi, samples, &nError ) ) {
					//WriteABFErr( nError, in_path, msgout );
					keep = false;
					break;
				}
				else {
					ABF_GetNumSamples( hFile, &abfFH, dwEpi, &iSampRead, &nError );
					for (unsigned int i=0; i<iSampRead; ++i) {
						lo = min(lo, samples[i]);
						hi = max(hi, samples[i]);
					}
				}
			}
			if ( keep ) {
				offset = 0.0;
				scaling = 32767.0 / max(abs(hi), abs(lo));
				scaling_ch.push_back(scaling);
				offset_ch.push_back(offset);

				insChan.insert("Channel");
				QUB_Tree chan = *insChan;
				++insChan;

				chan["Name"].setData( string(abfFH.sDACChannelName[0], 10) );
				chan["Units"].setData( string(abfFH.sDACChannelUnits[0], 8) );
				chan["Scaling"].setData(QTR_TYPE_DOUBLE, scaling);
				chan["Offset"].setData(QTR_TYPE_DOUBLE, offset);

				ixDigi.push_back( iCh );
				++iTotalChannels;
			}
		}
	}

	qdf["ADChannelCount"].setData(QTR_TYPE_INT, iTotalChannels);
	QUB_TreeIter insSeg = qdf["Segments"].children();

	// each epi add a segment
	for ( DWORD dwEpi=1; dwEpi<=dwMaxEpi; ++dwEpi ) {
		insSeg.insert("Segment");
		QUB_Tree seg = * insSeg;
		++insSeg;

		// seg["StartTime"].setData(QTR_TYPE_DOUBLE, ?);
		unsigned int N;
		ABF_GetNumSamples( hFile, &abfFH, dwEpi, &N, &nError );
		QUB_Tree channels = seg["Channels"];
		channels.setNumData(QTR_TYPE_SHORT, N, iTotalChannels, (void*) 0);
		short *ch = (short *) channels.data();

		for (int iCh=0; iCh<iChannels; ++iCh) {
			int jCh = abfFH.nADCSamplingSeq[iCh];
			if ( ! ABF_ReadChannel( hFile, &abfFH, jCh, dwEpi, samples, &iSampRead, &nError ) ) {
				WriteABFErr(nError, in_path, msgout);
				//return -5;
				memset(samples, 0, iSamples*sizeof(float));
			}
			scaling = scaling_ch[iCh];
			offset = offset_ch[iCh];
			for (unsigned int i=0; i<N; ++i)
				ch[i*iTotalChannels+iCh] = (int) floor((samples[i] - offset) * scaling + 0.5);
		}
		int iCha = iChannels;
		for (vector<int>::iterator itStim=ixDigi.begin(); itStim != ixDigi.end(); ++itStim, ++iCha) {
			if ( ! ABF_GetWaveform( hFile, &abfFH, *itStim, dwEpi, samples, &nError ) ) {
				WriteABFErr(nError, in_path, msgout);
				return -6;
			}
			scaling = scaling_ch[iCha];
			offset = offset_ch[iCha];
			for (unsigned int i=0; i<N; ++i)
				ch[i*iTotalChannels+iCha] = (int) floor((samples[i] - offset) * scaling + 0.5);
		}
		channels.unloadData();
		channels.setPreload(false);
	}

	ABF_Close( hFile, NULL );
	qdf.save();

	return 0; // success
}


QAX_API int   __stdcall  QAX_IsABFFile( const char *path )
{
	int nDataFormat, nError;
	return ABF_IsABFFile(path, &nDataFormat, &nError);
}

int WideCharCount(int16_t *s, int max_count)
{
	int i = 0;
	while ( (i < max_count) && s[i] )
		++i;
	return i;
}

void WideCharToQubtreeData(int16_t *s, int count, QUB_Tree node)
{
	count = min(count, WideCharCount(s, count)+1); // WCTMB wants terminating null too
	if ( count ) {
		int bytes = WideCharToMultiByte(CP_UTF8, 0, (LPCWSTR) s, count, NULL, 0, NULL, NULL);
		node.setData(QTR_TYPE_STRING, 1, bytes-1, (void*)0);
		node.loadData();
		WideCharToMultiByte(CP_UTF8, 0, (LPCWSTR) s, count, (LPSTR) node.data(), bytes, NULL, NULL);
	} else {
		node.clearData();
	}
}

string WideCharToStdString(int16_t *s, int count)
{
	count = min(count, WideCharCount(s, count)+1); // WCTMB wants terminating null too
	string out;
	if ( count ) {
		int bytes = WideCharToMultiByte(CP_UTF8, 0, (LPCWSTR) s, count, NULL, 0, NULL, NULL);
		out.resize(bytes-1);
		WideCharToMultiByte(CP_UTF8, 0, (LPCWSTR) s, count, (LPSTR) out.c_str(), bytes, NULL, NULL);
	}
	return out;
}

char NATURAL_PREFIX[] = {'?','?','?','p','n','u','m',' ','K','M','G','P','E', '?', '?'};
#define NP_OFF 7

#define CHUNK_SIZE QAX_COPYING_MAXBYTES

extern "C"
QAX_API int   __stdcall  Convert_Acquire_To_QDF( const char *in_path, const char *out_path,
											 int (__stdcall *write_msg)(const char *msg, void *data),
											 void *data )
{
	// wrap the text-output callback in a std::ostream:
	callbk_reportstream msgout(write_msg, data);

	int64_t channel_count, series_count, sweep_count, segment_count, data_count, frames_read;
	int64_t series_i, sweep_i, segment_i, channel_i, offset, chunk_size, i;
	acquirefile_read_series *series;
	acquirefile_read_sweep *sweep;
	acquirefile_read_segment *segment;
	acquirefile_read *file;
	int64_t total_count = 0;
	double sweep_start; // seconds since start of file

	file = acquirefile_read_open(in_path);
	if ( ! file )
		return -1;
	channel_count = file->header.analog_in_channel_count;
	series_count = file->header.series_count;

	// buffers
	short *buf = new short[CHUNK_SIZE*channel_count];
	vector<short*> chanbuf;
	for (channel_i=0; channel_i<channel_count; ++channel_i)
		chanbuf.push_back(buf + channel_i*CHUNK_SIZE);
	ostringstream notes;

	notes << "Acquire file version "
		  << file->prefix.version << '.' << file->prefix.minor_version
		  << endl << endl
		  << WideCharToStdString(file->header.comment, ACQUIRE_FILE_COMMENT_SIZE) << endl
		  << WideCharToStdString(file->header.technique, ACQUIRE_FILE_TECHNIQUE_SIZE)
		  << endl << endl;

	// setup qdf
	QUB_Tree qdf = QUB_Tree::Create("DataFile");
	if ( ! qdf.saveAs(out_path) ) {
		msgout << "Failed to write " << out_path << endl;
		acquirefile_read_close(file);
		delete [] buf;
		return -2;
	}
	double sampling = file->header.frame_interval * 1.0 / file->header.clock_rate;
	qdf["Sampling"].setData(QTR_TYPE_DOUBLE, sampling);
	qdf["Scaling"].setData(QTR_TYPE_DOUBLE, 1.0);
	qdf["ADDataSize"].setData(QTR_TYPE_INT, 2);
	qdf["ADDataType"].setData(QTR_TYPE_INT, (int) QTR_TYPE_SHORT);
	qdf["ADChannelCount"].setData(QTR_TYPE_INT, channel_count);

	// set up channels
	QUB_TreeIter insChan = qdf["SessionFile"]["DataChannels"].children();
	for (channel_i=0; channel_i<channel_count; ++channel_i) {
		insChan.insert("Channel");
		QUB_Tree chan = *insChan;
		++insChan;

		WideCharToQubtreeData(file->analog_in[channel_i].label, ACQUIRE_LABEL_SIZE, chan["Name"]);
		
		string units = WideCharToStdString(file->analog_in[channel_i].units, ACQUIRE_UNITS_SIZE);
		if ( file->analog_in[channel_i].natural_scale ) {
			ostringstream units_out;
			units_out << NATURAL_PREFIX[NP_OFF+file->analog_in[channel_i].natural_scale]
				      << units;
		    units = units_out.str();
		}
		chan["Units"].setData(units);
		AcquireSweepPrefixAnalogChannel *ch_prefix = file->series[0].sweeps[0]->analog + channel_i;
		chan["Scaling"].setData(QTR_TYPE_DOUBLE, 
			(ch_prefix->gain * pow(1e3, (double)file->analog_in[channel_i].natural_scale))
			* ((double) (ch_prefix->maximum_value - ch_prefix->minimum_value)
		       / (ch_prefix->maximum_voltage - ch_prefix->minimum_voltage)) );
		chan["Offset"].setData(QTR_TYPE_DOUBLE, 0.0);

		notes << "analog in channel " << channel_i << ": "
			  << chan["Name"].dataAsString()
			  << " (" << WideCharToStdString(file->analog_in[channel_i].signal, ACQUIRE_SIGNAL_SIZE)
			  << ") [" << chan["Units"].dataAsString() << "]" << endl;
	}
	qdf["SessionFile"]["Comment"].setData(notes.str());

	// set up lists (list series %i: sweep %j; series %i segments: sweep %j)
	QUB_TreeIter insList = qdf["SessionFile"]["Lists"].children();
	QUB_TreeIter insSweep, insSel;

	// copy segments
	QUB_TreeIter insSeg = qdf["Segments"].children();
	for (series_i=0; series_i<series_count; ++series_i) {
		series = file->series + series_i;
		sweep_count = series->descriptor.sweep_count;
		
		string comment = WideCharToStdString(series->descriptor.comment, ACQUIRE_COMMENT_SIZE);
		if ( comment.length() == 0 ) {
			ostringstream ost;
			ost << "Series " << (series_i+1);
			comment = ost.str();
		}
		insList.insert("List");
		insList->setData( comment );
		insSweep = insList->children();
		++insList;
		insList.insert("List");
		insList->setData( comment + " segments" );
		insSel = insList->children();
		++insList;

		for (sweep_i=0; sweep_i<sweep_count; ++sweep_i) {
			sweep = series->sweeps[sweep_i];
			sweep_start = 1e-7*(sweep->descriptor.start_time - file->header.date_time);
			comment = WideCharToStdString(sweep->descriptor.comment, ACQUIRE_COMMENT_SIZE);
			insSweep.insert(comment);
			insSweep->setNumData(QTR_TYPE_INT, 1, 2, (void*)0);
			insSweep->dataAs<int>(0, 0, 0) = total_count;
			segment_count = sweep->descriptor.segment_count;
			for (segment_i=0; segment_i<segment_count; ++segment_i) {
				segment = & sweep->segments[segment_i];
				data_count = segment->frames;

				insSel.insert("");
				insSel->setNumData(QTR_TYPE_INT, 1, 2, (void*)0);
				insSel->dataAs<int>(0, 0, 0) = total_count;
				total_count += data_count;
				insSel->dataAs<int>(0, 1, 0) = total_count - 1;
				++insSel;

				insSeg.insert("Segment");
				QUB_Tree seg = * insSeg;
				++insSeg;

				seg["StartTime"].setData(QTR_TYPE_DOUBLE, sweep_start + sampling*segment->descriptor.start);

				QUB_Tree channels = seg["Channels"];
				channels.setPreload(false);
				channels.setNumData(QTR_TYPE_SHORT, data_count, channel_count, (void*) 0);

				for (offset=0; offset<data_count; offset += CHUNK_SIZE) {
					chunk_size = data_count - offset;
					if ( chunk_size > CHUNK_SIZE )
						chunk_size = CHUNK_SIZE;
					for (channel_i=0; channel_i<channel_count; ++channel_i)
						frames_read = acquirefile_read_data_int16(file, series_i, sweep_i, segment_i, channel_i,
												offset, chunk_size, chanbuf[channel_i]);
					channels.loadData(offset, offset+chunk_size-1, false);
					short *chan_data = (short *) channels.data();
					for (i=0; i<chunk_size; ++i) {
						for (channel_i=0; channel_i<channel_count; ++channel_i)
							chan_data[i*channel_count + channel_i] = chanbuf[channel_i][i];
					}
					channels.unloadData();
				}
			}

			insSweep->dataAs<int>(0, 1, 0) = total_count - 1;
			++insSweep;
		}
	}
  
	acquirefile_read_close(file);
	delete [] buf;
	qdf.save();

	return 0; // success
}


